﻿#region License
// The PostgreSQL License
//
// Copyright (C) 2016 The Npgsql Development Team
//
// Permission to use, copy, modify, and distribute this software and its
// documentation for any purpose, without fee, and without a written
// agreement is hereby granted, provided that the above copyright notice
// and this paragraph and the following two paragraphs appear in all copies.
//
// IN NO EVENT SHALL THE NPGSQL DEVELOPMENT TEAM BE LIABLE TO ANY PARTY
// FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
// INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
// DOCUMENTATION, EVEN IF THE NPGSQL DEVELOPMENT TEAM HAS BEEN ADVISED OF
// THE POSSIBILITY OF SUCH DAMAGE.
//
// THE NPGSQL DEVELOPMENT TEAM SPECIFICALLY DISCLAIMS ANY WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS
// ON AN "AS IS" BASIS, AND THE NPGSQL DEVELOPMENT TEAM HAS NO OBLIGATIONS
// TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#endregion

using System.Linq;
using System.Linq.Expressions;
using System.Reflection;
using System.Text.RegularExpressions;
using JetBrains.Annotations;
using Microsoft.EntityFrameworkCore.Query.Expressions.Internal;

namespace Microsoft.EntityFrameworkCore.Query.ExpressionTranslators.Internal
{
    /// <summary>
    /// Translates Regex.IsMatch calls into PostgreSQL regex expressions for database-side processing.
    /// </summary>
    /// <remarks>
    /// http://www.postgresql.org/docs/current/static/functions-matching.html
    /// </remarks>
    public class NpgsqlRegexIsMatchTranslator : IMethodCallTranslator
    {
        static readonly MethodInfo IsMatch;
        static readonly MethodInfo IsMatchWithRegexOptions;

        const RegexOptions UnsupportedRegexOptions = RegexOptions.RightToLeft | RegexOptions.ECMAScript;

        static NpgsqlRegexIsMatchTranslator()
        {
            IsMatch = typeof (Regex).GetTypeInfo().GetDeclaredMethods("IsMatch").Single(m =>
                m.GetParameters().Count() == 2 &&
                m.GetParameters().All(p => p.ParameterType == typeof(string))
            );
            IsMatchWithRegexOptions = typeof(Regex).GetTypeInfo().GetDeclaredMethods("IsMatch").Single(m =>
               m.GetParameters().Count() == 3 &&
               m.GetParameters().Take(2).All(p => p.ParameterType == typeof(string)) &&
               m.GetParameters()[2].ParameterType == typeof(RegexOptions)
            );
        }

        public Expression Translate([NotNull] MethodCallExpression methodCallExpression)
        {
            // Regex.IsMatch(string, string)
            if (methodCallExpression.Method == IsMatch)
            {
                return new RegexMatchExpression(
                    methodCallExpression.Arguments[0],
                    methodCallExpression.Arguments[1],
                    RegexOptions.None
                );
            }

            // Regex.IsMatch(string, string, RegexOptions)
            if (methodCallExpression.Method == IsMatchWithRegexOptions)
            {
                var constantExpr = methodCallExpression.Arguments[2] as ConstantExpression;

                if (constantExpr == null)
                {
                    return null;
                }

                var options = (RegexOptions) constantExpr.Value;

                if ((options & UnsupportedRegexOptions) != 0)
                {
                    return null;
                }

                return new RegexMatchExpression(
                    methodCallExpression.Arguments[0],
                    methodCallExpression.Arguments[1],
                    options
                );
            }

            return null;
        }
    }
}
